#!/usr/bin/env python

# exit codes
OSF_RETRY = 2
NOT_OK = 1
OK = 0

# modules for OPUS internal poller
import sys                  # access to command-line args
import string               # string functions
import opusutil             # utilites like PrintMsg
import re                   # regular expression handling
import traceback            # printing of stack traces 

# python modules for this application
import os
import getopt
import time
import glob

# local modules for this application
import base_poller
import siname
import cdbsquery
import hst_missionquery

# exceptions
class MissingIORonCommandLine(Exception):
  pass

class NarrowOfOrbObjectFailed(Exception):
  pass

class FailedToRegisterPipelineApp(Exception):
  pass

class UsageError(Exception):
  pass

class NotFitsOrDbInput(Exception):
  pass

class MissingRequiredFiles(Exception):
  pass

class FailedResourceInit(Exception):
  pass

class PollingFailed(Exception):
  pass

class HaltEvent(Exception):
  pass

class DatasetNotFoundInOSF(Exception):
  pass

class DBException(Exception):
  pass

class NoFilesFound(Exception):
  pass

class ArchivingIncomplete(Exception):
  pass

class CantFindGeneralAvailabilityDate(Exception):
  pass

class Cdbs_feedback(base_poller.InternalPoller):
  """
########################################################################
Class: Cdbs_feedback

Description:
------------
The class for interactive or pipeline processing of the CDBS feedback
mechanism for a CDBS reference file delivery.

Members:
--------
trlpath = directory for trailer file to be updated
cdbs_db  = database object for accessing CDBS tables
reffile_db = database object for accessing REFFILE tables
opus_db = database object for accessing OPUS tables

Methods:
--------
usage() - prints an interactive usage message and raises an exception

########################################################################
  """
  def __init__(self):
    # initialize members
    self._inpath = None
    self._trlpath = None
    self._update_limit = 5000 # number of rows in each DB update chunk
    self._cdbs_db  = None
    self._reffile_db = None
    self._opus_db  = None

  def initialize(self):
    #
    # call superclass constructor
    base_poller.InternalPoller.__init__(self)

  #--------------------------------------------------------------------------
  def usage(self):
    opusutil.PrintMsg('F','Interactive Usage: cdbfb.py [-h] -d delivery_name -p inpath')
    raise UsageError

  #--------------------------------------------------------------------------
  #
  # Name: interactive
  #
  # Description: Performs command line parsing and set-up for
  #              processing in interactive mode.  Runs the
  #              task for the command-line arguments given.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 04/14/04 48141  MSwam     first version
  # 07/15/10 57271  MSwam     remove db retry parameter
  # 04/20/11 67900  MSwam     let do_cdbfb return exception to caller
  #
  #--------------------------------------------------------------------------
  def interactive(self):
    #
    # parse command line arguments
    try:
      opts, args = getopt.getopt(sys.argv[1:], "hd:p:", 
                   ["help","delivery_name=","inpath="])
    except getopt.GetoptError:
      # print help information and raise exception
      self.usage()

    # initialize
    delivery_name = None

    # parse options
    for flag, value in opts:
      if flag in ("-h", "--help"):
        self.usage()
      if flag in ("-d", "--delivery_name"):
        delivery_name = value
      if flag in ("-p", "--inpath"):
        self._inpath = value

    if delivery_name == None:
      # no command line dataset supplied, so try to pick up OSF_DATASET env var
      try:
        delivery_name = os.environ['OSF_DATASET']
        opusutil.PrintMsg("D","found OSF_DATASET="+delivery_name)
      except KeyError:
        opusutil.PrintMsg("E",
          "No interactive arguments found AND FAILED to find OSF_DATASET")
        self.usage()

    if self._inpath == None:
      #  no command line input dir specified, so look for env var
      try:
        self._inpath = os.environ['INPATH']
        opusutil.PrintMsg("D","INPATH="+self._inpath)
      except KeyError:
        # set the directory for file operations to the current directory
        self._inpath = os.getcwd()

    # set the directory/flag for trailer file use
    using_trl = 0
    try:
      self._trlpath = os.environ['TRLPATH']
      using_trl = 1
    except KeyError:
      # default to current directory
      self._trlpath = self._inpath
    opusutil.PrintMsg("D","TRLPATH="+self._trlpath)

    # connect to databases
    self._cdbs_db = cdbsquery.Cdbs_db(os.environ['CDBS_SERVER'], 
                                      os.environ['CDBS_DB'])
    
    self._reffile_db = cdbsquery.Cdbs_db(os.environ['REFFILE_SERVER'], 
                                         os.environ['REFFILE_DB'])
    
    self._opus_db = hst_missionquery.HST_Mission_db(os.environ['OPUS_SERVER'], 
                                                    os.environ['OPUS_DB'])

    # run the computations
    if using_trl:
       opusutil.OpenTrl(self._trlpath + os.sep + delivery_name + ".trl")
       opusutil.PrintMsg("I","----- FB started for: "+delivery_name+" -----")

    self.do_cdbfb(delivery_name)

    if using_trl:
       opusutil.PrintMsg("I","----- FB ended for: "+delivery_name+" -----")
       opusutil.CloseTrl()

    # close the database connections
    self._cdbs_db.close()
    self._reffile_db.close()
    self._opus_db.close()

    return

  #-------------------------------------------------------------------------
  # 
  # Name: pipeline_init
  #
  # Description: Initialization for pipeline mode, including reading
  #              XML file of reference file specs and resource file
  #              parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/14/04 48141  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def pipeline_init(self):
    # obtain resource items in a dictionary and assign to globals
    #
    try:
      resource_items = base_poller.InternalPoller.get_resource_items(self,
                            ["INPATH",
                             "OPUS_SERVER",
                             "OPUS_DB",
                             "CDBS_DB",
                             "CDBS_SERVER"] )
    except:
      opusutil.PrintMsg("E","failed to get required resource entry")
      raise
    #
    opusutil.PrintMsg("I","Resource items read.")
    self._inpath = resource_items['INPATH']
    # trailer is assumed to be in same directory as catalog
    self._trlpath = self._inpath
    #
    # connect to the databases
    self._cdbs_db = cdbsquery.Cdbs_db(resource_items['CDBS_SERVER'], 
                                      resource_items['CDBS_DB'])
    self._reffile_db = cdbsquery.Cdbs_db(resource_items['REFFILE_SERVER'], 
                                         resource_items['REFFILE_DB'])
    self._opus_db = hst_missionquery.HST_Mission_db(resource_items['OPUS_SERVER'], 
                                      resource_items['OPUS_DB'])
    return

  #-------------------------------------------------------------------------
  #
  # Name: verifyDisk
  #
  # Description: For a given dataset, verifies that files exist on
  #              the corresponding reference file disk.
  #              It is assumed that the instrument for which the
  #              reference file applies can be inferred from the reference
  #              file name.  The instrument name is then converted into
  #              an environment variable name that specifies the location
  #              of the reference files for that instrument.  A file search
  #              is performed to confirmed that at least one file with the
  #              given rootname exists in the instrument-specific
  #              reference file directory.
  #
  # Input: dataset - rootname of the reference file
  #
  # Returns: 1 - if files are found on disk
  #
  # Exceptions: NoFilesFound - raised if no reference files are found on disk
  #                            for the given rootname
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/14/04 48141  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def verifyDisk(self, dataset):
    # determine instrument from dataset name
    inst = siname.WhichCDBSInstrument(dataset)
    
    # convert to environment variable name 
    # (fudge IRAF variable, oref$, to UNIX variable, $oref)
    envvar_name = siname.add_IRAF_prefix(inst)
    envvar_name = "$"+envvar_name[:-1]

    # look for at least one qualifying ref file on disk
    filespec = os.path.expandvars(envvar_name + os.sep + 
                                  string.lower(dataset) + "*")
    flist = glob.glob(filespec)
    for f in flist:
      opusutil.PrintMsg("I","found "+str(f))
      return 1
    #
    # no files found, raise exception
    opusutil.PrintMsg("E","Failed to find ref file using "+filespec)
    raise NoFilesFound

  #-------------------------------------------------------------------------
  #
  # Name: verifyArchive
  #
  # Description: For a given dataset, verifies that files have been archived
  #              by checking for a valid ingest indication in the OPUS_DB
  #              table dads_archive.
  #
  # Input: dataset - rootname of the reference file
  #
  # Returns: 1 - if ingest response is OK
  #
  # Exceptions: ArchivingIncomplete - raised if a dads_archive record is 
  #                           not found or does not indicate ingest success
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/12/04 48141  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def verifyArchive(self, dataset):
    #
    # query dads_archive table for response value
    response = self._opus_db.dads_archive_response(dataset)
    if response != "OK":
      opusutil.PrintMsg("W","Archiving incomplete for "+dataset+": response="+
                        str(response))
      raise ArchivingIncomplete
    #
    # success
    return 1

  #-------------------------------------------------------------------------
  #
  # Name: catalogInserts
  #
  # Description: Returns a list of dataset names read from a CDBS
  #              release catalog.  Only catalog INSERTION (I) entries
  #              are returned (DELETION (D) entries are ignored).
  #
  # Input: delivery_name - name of the delivery to process
  #
  # Output: datasets - list of insertion datasets in a CDBS release catalog
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/14/04 48141  MSwam     first version
  # 05/04/05 51433  MSwam     remove extension trim (caller must handle)
  #
  #-------------------------------------------------------------------------
  def catalogInserts(self, delivery_name):
    # form catalog name from delivery name
    catalog_name = self._inpath + os.sep + delivery_name + '.cat'

    # read catalog insert datasets into a list
    datasets = []
    try:
      for line in open(catalog_name, 'r').readlines():
        # split line and save dataset if entry is an insertion
        parts = string.split(line)
        if parts[1] == "I":
          datasets.append(string.lower(parts[0]))
    except IOError:
      opusutil.PrintMsg("E","Failed to read catalog "+catalog_name)
      raise
    #
    return datasets

  #-------------------------------------------------------------------------
  #
  # Name: applyCDBSsql
  #
  # Description: Reads SQL from the CDBS sql input file and applies
  #              it against the REFFILE_DB.  The input file consists
  #              of a set of multi-line INSERT statements, each followed
  #              by a "go" directive.  This routine will use the "go" entry
  #              to determine when a particular query has been completely
  #              read, and then apply that query to the database, and read on
  #              for more.
  #              
  # Input: delivery_number - name of the delivery to process
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/26/10 49175  MSwam     first version
  # 05/19/11 68398  MSwam     allow GO or go (case-insensitive)
  #
  #-------------------------------------------------------------------------
  def applyCDBSsql(self, delivery_number):
    # form SQL file name from delivery name (cdbs_deliveryNNNNN.sql)
    sqlfile_name = self._inpath + os.sep + 'cdbs_delivery'+str(delivery_number)+ '.sql'

    # read SQL query text until a "go" is encountered, then apply the query
    # and continue reading for the next query (used to support GO as well)
    try:
      querytxt = ''
      skip_dblogin = 1
      for line in open(sqlfile_name, 'r').readlines():
        if len(line)==3 and (line[0:2]=='GO' or line[0:2]=='go'):
           if skip_dblogin:
              # we haven't yet skipped over the DB login lines; now we HAVE
              opusutil.PrintMsg("I","Skipping over DB login...")
              skip_dblogin = 0
              querytxt = ''
              continue
           # query text complete, apply to DB, and clear for next query
           self._reffile_db.update(querytxt)
           querytxt = ''
        else:
           # continue to build the next query
           querytxt += line
    except IOError:
      opusutil.PrintMsg("E","Failed to read SQL file: "+sqlfile_name)
      raise
    #
    opusutil.PrintMsg("I","SQL updates applied to REFFILE_DB ")
    return

  #-------------------------------------------------------------------------
  # 
  # Name: do_cdbfb
  #
  # Description: Verifies and delivers the feedback response to the
  #              CDBS system indicating successful installation of a
  #              CDBS delivery in the pipeline.
  #
  # Input:
  # delivery_name - name of the delivery to process
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/14/04 48141  MSwam     first version
  # 05/04/05 51433  MSwam     trim extension before verifyArchive call
  # 08/31/10 65923  MSwam     added ga_date update, since it is not in the SQL
  # 01/03/11 67112  MSwam     add DateTrim to remove subseconds that mess up
  #                             update call under MSSQL
  #-------------------------------------------------------------------------
  def do_cdbfb(self, delivery_name):
    # obtain list of ref file "insert" datasets from catalog file
    datasets = self.catalogInserts(delivery_name)
    opusutil.PrintMsg("D","datasets="+str(datasets))

    # loop through datasets to verify files on disk (exception raised if failed)
    for dataset in datasets:
      self.verifyDisk(dataset)

    # Disk Verification complete, now parse delivery number
    parts = string.split(delivery_name,"_")
    delivery_number = int(parts[1])

    # get todays date-time
    load_date = time.strftime("%m/%d/%Y %H:%M:%S",time.localtime(time.time()))
    opusutil.PrintMsg("I",delivery_name + " opus_load_date = " +
                      load_date)

    # create a query for each CDBS-supported instrument, since we don't
    # know which ones are part of the delivery
                                              # syntax to add to a tuple
    cdbs_instruments = siname.CDBS_supports + ('SYNPHOT',) 
    for i in cdbs_instruments:
      querytxt = ("UPDATE " + siname.get_cdbs_prefix(i) + 
                  "_file SET opus_load_date = '" + load_date + "' " +
                  " WHERE delivery_number = " + str(delivery_number))
      self._cdbs_db.update(querytxt)

    # loop through the datasets again to verify files were archived
    for dataset in datasets:
      # trim "extension" from name before lookup, where extension is
      # the last underscore and any trailing characters
      last_underscore = dataset.rfind("_")
      if last_underscore > -1:
         self.verifyArchive(dataset[:last_underscore])
      else:
         # no underscore to trim from
         self.verifyArchive(dataset)

    # Archive Verification complete
    opusutil.PrintMsg("I",delivery_name + " archive_date = " +
                      load_date)

    # create a query for each CDBS-supported instrument, since we don't
    # know which ones are part of the delivery
    for i in cdbs_instruments:
      querytxt = ("UPDATE " + siname.get_cdbs_prefix(i) + 
                  "_file SET archive_date = '" + load_date + "' " +
                  " WHERE delivery_number = " + str(delivery_number))
      self._cdbs_db.update(querytxt)

    # obtain the general_availability_date from the delivery so it can
    # be added to the REFFILE_DB entries (it doesn't come across in the
    # delivery SQL)
    ga_date = None
    for i in cdbs_instruments:
      # we don't know which instrument, so issue queries until a match occurs
      querytxt = ("SELECT DISTINCT general_availability_date FROM " + 
                  siname.get_cdbs_prefix(i) + "_file WHERE delivery_number = "+
                  str(delivery_number))
      result = [[]]
      self._cdbs_db.zombie_select(querytxt, result)
      if len(result) > 0:
         # we found the appropriate instrument, save date and break out
         ga_date = opusutil.DateTrim(str(result[0][0]))
         opusutil.PrintMsg('I','general_availability_date found: '+str(ga_date))
         break
    if ga_date == None:
         opusutil.PrintMsg('F','FAILED to find general_availability_date for '+
                               ' CDBS delivery '+str(delivery_number))
         raise CantFindGeneralAvailabilityDate

    # now that the CDBS_DB is updated, apply delivery SQL to the REFFILE_DB
    # so that the pipeline can start using the new ref files
    #
    self.applyCDBSsql(delivery_number)

    # also need to set the opus_load_date, archive_date, and
    # general_availability_date in the REFFILE_DB
    # to values in the CDBS_DB
    #
    # (again create a query for each CDBS-supported instrument, since we don't
    #  know which ones are part of the delivery)
    for i in cdbs_instruments:
      querytxt = ("UPDATE " + siname.get_cdbs_prefix(i) + 
                  "_file SET archive_date = '" + load_date + 
                  "', opus_load_date = '" + load_date + "', " +
                  "general_availability_date = '" + str(ga_date) + "' " +
                  " WHERE delivery_number = " + str(delivery_number))
      opusutil.PrintMsg("I","ACTIVATING CDBS delivery "+str(delivery_number)+
                            " for pipeline use using query="+querytxt)
      self._reffile_db.update(querytxt)

    return

  #-------------------------------------------------------------------------
  # Name: processHaltEvent
  #
  # Description: Handle a HALT event (override base class).
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/22/04 51879  MSwam     use base poller class
  #
  #-------------------------------------------------------------------------
  def processHaltEvent(self, event):
    #
    opusutil.PrintMsg("I","Received HALT event. Stopping.")
    self._cdbs_db.close()
    self._reffile_db.close()
    self._opus_db.close()
    raise HaltEvent

  #-------------------------------------------------------------------------
  # Name: processOsfEvent
  #
  # Description: Handle OSF events in pipeline mode.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/01/04 49754  MSwam     first version
  # 05/11/04 49754  MSwam     fix success trigger name
  # 08/06/04 49860  Sontag    Changed an IDL field name
  # 09/22/04 51879  MSwam     Use base_poller class
  # 10/29/04 52310  MSwam     remove db closes (only HALT event should)
  # 04/04/05 51433  MSwam     remove traceback from ArchivingIncomplete case
  #
  #-------------------------------------------------------------------------
  def processOsfEvent(self, event):
        # get trigger info
        dataset, data_id = self.getDatasetDataidFromOsfEvent(event)

        try:
          opusutil.OpenTrl(self._trlpath + os.sep + dataset + ".trl")
        except:
          # any exception marks the OSF with OSF_ERROR
          traceback.print_exc()
          #
          # an error occurred during trailer opening, set the OSF to failed
          try:
            opusutil.PrintMsg("E","closing event with OSF_ERROR")
            self._papp.closeEvent("OSF_ERROR", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_ERROR event")
          return

        opusutil.PrintMsg("I","----- CDBFB started for: "+dataset+" -----")
        try:
          self.do_cdbfb(dataset)
        #
        except ArchivingIncomplete:
          #
          # the archiving is not yet done, exit with RETRY status
          try:
            opusutil.PrintMsg("W","closing event with OSF_RETRY")
            self._papp.closeEvent("OSF_RETRY", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_RETRY event")
          opusutil.PrintMsg("I","----- CDBFB ended for: "+dataset+" -----")
          opusutil.CloseTrl()
          return
        except:
          # any other exception marks the OSF with OSF_ERROR
          traceback.print_exc()
          #
          # an error occurred during processing, set the OSF
          try:
            opusutil.PrintMsg("E","closing event with OSF_ERROR")
            self._papp.closeEvent("OSF_ERROR", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_ERROR event")
          opusutil.PrintMsg("I","----- CDBFB ended for: "+dataset+" -----")
          opusutil.CloseTrl()
          return

        try:
          self._papp.closeEvent("OSF_NORMAL", event.eid)
        except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_NORMAL event")
        opusutil.PrintMsg("I","----- CDBFB ended for: "+dataset+" -----")
        opusutil.CloseTrl()

#-------------------------------------------------------------------------
# The main routine. Handles interactive and pipeline mode.
#
# History:
# -------- -----  --------  --------------------------------------
# 02/28/05 51433  MSwam     use base_poller base class
# 04/20/11 67900  MSwam     handle ArchivingIncomplete exception in interactive mode
#
#-------------------------------------------------------------------------
#
# initialize
try:
  poller = Cdbs_feedback()
  poller.initialize()
  poller.pipeline_init()
#
except base_poller.InteractiveMode:
  opusutil.PrintMsg("I","Running command-line mode.")
  try:
    poller.interactive()
  except ArchivingIncomplete:
    opusutil.PrintMsg("W","archiving not complete, exiting with OSF_RETRY...")
    sys.exit(OSF_RETRY)
  except UsageError:
    sys.exit(NOT_OK)
  except:
    traceback.print_exc()
    opusutil.PrintMsg("F","Command-line mode Failed.")
    sys.exit(NOT_OK)
  sys.exit(OK)
#
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","initialization failed")
  sys.exit(NOT_OK)

opusutil.PrintMsg("I","Pipeline set-up complete.")
try:
  poller.run()
except HaltEvent:
  sys.exit(OK)
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","pipeline_process failed")
  sys.exit(NOT_OK)
